@@ -20,6 +20,7 @@ from alphafold2_pytorch.rotary import DepthWiseConv1d, AxialRotaryEmbedding, Fix
 from se3_transformer_pytorch import SE3Transformer
 from se3_transformer_pytorch.utils import torch_default_dtype, fourier_encode
 from en_transformer import EnTransformer
+from egnn_pytorch import EGNN_Network
 
 from performer_pytorch import FastAttention, ProjectionUpdater
 
@@ -32,6 +33,13 @@ Logits = namedtuple('Logits', ['distance', 'theta', 'phi', 'omega'])
 def exists(val):
     return val is not None
 
+def maybe(fn):
+    def inner(t, *args, **kwargs):
+        if not exists(t):
+            return None
+        return fn(t, *args, **kwargs)
+    return inner
+
 def default(val, d):
     if exists(val):
         return val
@@ -457,6 +465,7 @@ class LinearAttention(Attention):
             rot_q, rot_k = cast_tuple(rotary_emb, 2)
             q = apply_rotary_pos_emb(q, rot_q)
             k = apply_rotary_pos_emb(k, rot_k)
+            v = apply_rotary_pos_emb(v, rot_k)
 
         # linear attention masking
 
@@ -522,6 +531,7 @@ class SparseAttention(Attention):
             rot_q, rot_k = cast_tuple(rotary_emb, 2)
             q = apply_rotary_pos_emb(q, rot_q)
             k = apply_rotary_pos_emb(k, rot_k)
+            v = apply_rotary_pos_emb(v, rot_k)
 
         key_pad_mask = None
         if exists(mask):
@@ -699,6 +709,7 @@ class CoordModuleMDS(nn.Module):
     def forward(
         self,
         *,
+        seq,
         distance_pred,
         trunk_embeds,
         N_mask,
@@ -706,7 +717,7 @@ class CoordModuleMDS(nn.Module):
         C_mask,
         cloud_mask,
         bb_flat_mask_crossed,
-        num_backbone_atoms
+        atom_mask
     ):
         if self.predict_real_value_distances:
             distances, distance_std = distance_pred.unbind(dim = -1)
@@ -733,7 +744,7 @@ class CoordModuleMDS(nn.Module):
 
         coords = rearrange(coords_3d, 'b c n -> b n c')
         # will init all sidechain coords to cbeta if present else c_alpha
-        coords = sidechain_container(coords, n_aa = num_backbone_atoms, cloud_mask = cloud_mask)
+        coords = sidechain_container(seq, coords, atom_mask = atom_mask, cloud_mask = cloud_mask)
         coords = rearrange(coords, 'b n l d -> b (n l) d')
 
         return coords
@@ -803,15 +814,17 @@ class Alphafold2(nn.Module):
         cross_attn_compress_ratio = 1,
         msa_tie_row_attn = False,
         template_attn_depth = 2,
-        num_backbone_atoms = 1,                # number of atoms to reconstitute each residue to, defaults to 3 for C, C-alpha, N
+        atoms = 'backbone-only',               # number of atoms to reconstitute each residue to, defaults to 3 for C, C-alpha, N
         predict_angles = False,
         symmetrize_omega = False,
         predict_coords = False,                # structure module related keyword arguments below
+        refine_coords = False,                 # don't start refiners if false
         predict_real_value_distances = False,
         trunk_embeds_to_se3_edges = 0,         # feeds pairwise projected logits from the trunk embeddings into the equivariant transformer as edges
         se3_edges_fourier_encodings = 4,       # number of fourier encodings for se3 edges
         return_aux_logits = False,
-        use_se3_transformer = True,            # uses SE3 Transformer - but if set to false, will use the new E(n)-Transformer
+        template_embedder_type = 'en',         # use E(n) Transformer for embedding templates
+        structure_module_type = 'se3',         # uses SE3 Transformer - but if set to false, will use the new E(n)-Transformer
         structure_module_dim = 4,
         structure_module_depth = 1,
         structure_module_heads = 1,
@@ -819,6 +832,7 @@ class Alphafold2(nn.Module):
         structure_module_refinement_iters = 2,
         structure_module_knn = 2,
         structure_module_adj_neighbors = 2,
+        structure_module_adj_dim = 4,
         cross_attn_linear = False,
         cross_attn_linear_projection_update_every = 1000,
         cross_attn_kron_primary = False,
@@ -831,8 +845,6 @@ class Alphafold2(nn.Module):
         coords_module = None                    # custom coords generation module
     ):
         super().__init__()
-        assert num_backbone_atoms in {1, 3, 4}, 'must be either residue level, or reconstitute to atomic coordinates of 3 for the C, Ca, N of backbone, or 4 of C-beta as well'
-
         self.dim = dim
 
         # token embedding
@@ -867,33 +879,36 @@ class Alphafold2(nn.Module):
 
         self.return_aux_logits = return_aux_logits
 
-        # template sidechain encoding
+        # template sidechain encoding - only if needed
 
-        self.use_se3_transformer = use_se3_transformer
+        self.template_embedder_type = template_embedder_type
 
-        if use_se3_transformer:
-            self.template_sidechain_emb = SE3TemplateEmbedder(
-                dim = dim,
-                dim_head = dim,
-                heads = 1,
-                num_neighbors = 12,
-                depth = 4,
-                input_degrees = 2,
-                num_degrees = 2,
-                output_degrees = 1,
-                reversible = True,
-                tie_key_values = True,
-                one_headed_key_values = True,
-                num_positions = max_seq_len
-            )
-        else:
-            self.template_sidechain_emb = EnTransformer(
-                dim = dim,
-                dim_head = dim,
-                heads = 1,
-                neighbors = 32,
-                depth = 4
-            )
+        if isinstance(template_embedder_type, str): 
+            if template_embedder_type == 'se3':
+                self.template_sidechain_emb = SE3TemplateEmbedder(
+                    dim = dim,
+                    dim_head = dim,
+                    heads = 1,
+                    num_neighbors = 12,
+                    depth = 4,
+                    input_degrees = 2,
+                    num_degrees = 2,
+                    output_degrees = 1,
+                    reversible = True,
+                    tie_key_values = True,
+                    one_headed_key_values = True,
+                    num_positions = max_seq_len
+                )
+            elif template_embedder_type == 'en':
+                self.template_sidechain_emb = EnTransformer(
+                    dim = dim,
+                    dim_head = dim,
+                    heads = 1,
+                    neighbors = 32,
+                    depth = 4
+                )
+            else:
+                raise ValueError('template embedder type must be either "se3" or "en"')
 
         # custom embedding projection
 
@@ -970,10 +985,33 @@ class Alphafold2(nn.Module):
         if self.has_linear_attn:
             self.proj_updater = ProjectionUpdater(self.net, cross_attn_linear_projection_update_every)
 
+        # atom masking
+
+        if atoms == 'backbone-only':
+            atom_mask = torch.tensor([1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
+        elif atoms == 'backbone-with-cbeta':
+            atom_mask = torch.tensor([1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
+        elif atoms == 'all':
+            atom_mask = torch.tensor([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
+        elif atoms == 'backbone-with-oxygen':
+            atom_mask = torch.tensor([1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
+        elif atoms == 'backbone-with-cbeta-and-oxygen':
+            atom_mask = torch.tensor([1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
+        elif torch.is_tensor(atoms):
+            atom_mask = atoms
+        else:
+            raise ValueError('atoms needs to be a valid string or a mask tensor of shape (14,) ')
+
+        self.num_atoms = atom_mask.sum(-1).item()
+
+        assert tuple(atom_mask.shape) == (constants.NUM_COORDS_PER_RES,), 'atoms needs to be of the correct shape (14,)'
+
+        self.register_buffer('atom_mask', atom_mask.bool())
+
         # to distogram output
 
-        self.num_backbone_atoms = num_backbone_atoms
-        needs_upsample = num_backbone_atoms > 1
+        trunk_upsample_factor = self.num_atoms
+        needs_upsample = trunk_upsample_factor > 1
 
         self.predict_real_value_distances = predict_real_value_distances
         dim_distance_pred = constants.DISTOGRAM_BUCKETS if not predict_real_value_distances else 2   # 2 for predicting mean and standard deviation values of real-value distance
@@ -981,9 +1019,9 @@ class Alphafold2(nn.Module):
         self.to_distogram_logits = nn.Sequential(
             nn.LayerNorm(dim),
             nn.Sequential(
-                nn.Linear(dim, dim * (num_backbone_atoms ** 2)),
+                nn.Linear(dim, dim * (trunk_upsample_factor ** 2)),
                 Rearrange('b h w c -> b c h w'),
-                nn.PixelShuffle(num_backbone_atoms),
+                nn.PixelShuffle(trunk_upsample_factor),
                 Rearrange('b c h w -> b h w c')
             ) if needs_upsample else nn.Identity(),
             nn.Linear(dim, dim_distance_pred)
@@ -1008,13 +1046,14 @@ class Alphafold2(nn.Module):
         self.structure_num_global_nodes = structure_num_global_nodes
 
         if structure_num_global_nodes > 0:
-            assert use_se3_transformer, 'se3 transformer must be used in order to use global node token feature'
+            assert structure_module_type == 'se3', 'se3 transformer must be used in order to use global node token feature'
             self.global_queries = nn.Parameter(torch.randn(structure_num_global_nodes, dim))
             self.global_pool_attns.append(Attention(dim = dim))
             global_feats_dim = dim
 
         # to coordinate output
 
+        self.refine_coords = refine_coords
         self.predict_coords = predict_coords
         self.mds_iters = mds_iters
         self.structure_module_refinement_iters = structure_module_refinement_iters
@@ -1027,12 +1066,12 @@ class Alphafold2(nn.Module):
 
         self.to_equivariant_net_edges = nn.Linear(dim, trunk_embeds_to_se3_edges) if trunk_embeds_to_se3_edges > 0 else None
 
-        if self.predict_coords:
+        if self.refine_coords:
             with torch_default_dtype(torch.float64):
                 self.structure_module_embeds = nn.Embedding(num_tokens, structure_module_dim)
                 self.atom_tokens_embed = nn.Embedding(len(ATOM_IDS), structure_module_dim)
 
-                if use_se3_transformer:
+                if structure_module_type == 'se3':
                     self.structure_module = SE3TransformerWrapper(
                         dim = structure_module_dim,
                         depth = structure_module_depth,
@@ -1045,13 +1084,13 @@ class Alphafold2(nn.Module):
                         attend_sparse_neighbors = True,
                         edge_dim = edge_dim,
                         num_adj_degrees = structure_module_adj_neighbors,
-                        adj_dim = 4,
+                        adj_dim = structure_module_adj_dim,
                         global_feats_dim = global_feats_dim,
                         tie_key_values = True,
                         one_headed_key_values = True,
                         num_positions = max_seq_len * 14 # hard code as 14 since residual to atom is not flexible atm
                     )
-                else:
+                elif structure_module_type == 'en':
                     self.structure_module = EnTransformer(
                         dim = structure_module_dim,
                         depth = structure_module_depth,
@@ -1063,6 +1102,18 @@ class Alphafold2(nn.Module):
                         num_adj_degrees = structure_module_adj_neighbors,
                         adj_dim = 4
                     )
+                elif structure_module_type == 'egnn':
+                    self.structure_module = EGNN_Network(
+                        dim = structure_module_dim,
+                        depth = structure_module_depth,
+                        num_positions = max_seq_len * 14, # hard code as 14 since residual to atom is not flexible atm
+                        edge_dim = edge_dim,
+                        num_adj_degrees = structure_module_adj_neighbors,
+                        adj_dim = structure_module_adj_dim,
+                        coor_weights_clamp_value = 2.
+                    )
+                else:
+                    raise ValueError('structure module must be either "se3", "en", or "egnn" for SE3 Transformers, E(n)-Transformers, or EGNN respectively')
 
         # aux confidence measure
         self.lddt_linear = nn.Linear(structure_module_dim, 1)
@@ -1087,7 +1138,8 @@ class Alphafold2(nn.Module):
         templates_sidechains = None,
         embedds = None,
         return_trunk = False,
-        return_confidence = False
+        return_confidence = False,
+        refine = True
     ):
         assert not (self.disable_token_embed and not exists(seq_embed)), 'sequence embedding must be supplied if one has disabled token embedding'
         assert not (self.disable_token_embed and not exists(msa_embed)), 'msa embedding must be supplied if one has disabled token embedding'
@@ -1166,14 +1218,14 @@ class Alphafold2(nn.Module):
             # todo (make efficient)
 
             if exists(templates_sidechains):
-                if self.use_se3_transformer:
+                if self.template_embedder_type == 'se3':
                     t_seq = self.template_sidechain_emb(
                         t_seq,
                         templates_sidechains,
                         templates_coors,
                         mask = templates_mask
                     )
-                else:
+                elif self.template_embedder_type == 'en':
                     shape = t_seq.shape
                     t_seq = rearrange(t_seq, 'b t n d -> (b t) n d')
                     templates_coors = rearrange(templates_coors, 'b t n c -> (b t) n c')
@@ -1282,19 +1334,18 @@ class Alphafold2(nn.Module):
 
         # prepare mask for backbone coordinates
 
-        assert self.num_backbone_atoms > 1, 'must constitute to at least 3 atomic coordinates for backbone'
-
-        N_mask, CA_mask, C_mask = scn_backbone_mask(seq, boolean = True, n_aa = self.num_backbone_atoms)
+        N_mask, CA_mask, C_mask = scn_backbone_mask(seq, boolean = True, n_aa = self.num_atoms)
 
         cloud_mask = scn_cloud_mask(seq, boolean = True)
         flat_cloud_mask = rearrange(cloud_mask, 'b l c -> b (l c)')
         chain_mask = (mask.unsqueeze(-1) * cloud_mask)
         flat_chain_mask = rearrange(chain_mask, 'b l c -> b (l c)')
 
-        bb_flat_mask = rearrange(chain_mask[..., :self.num_backbone_atoms], 'b l c -> b (l c)')
+        bb_flat_mask = rearrange(chain_mask[..., :self.num_atoms], 'b l c -> b (l c)')
         bb_flat_mask_crossed = rearrange(bb_flat_mask, 'b i -> b i ()') * rearrange(bb_flat_mask, 'b j -> b () j')
 
         coords = self.trunk_to_coords(
+            seq = seq, 
             distance_pred = distance_pred,
             trunk_embeds = trunk_embeds,
             N_mask = N_mask,
@@ -1302,17 +1353,18 @@ class Alphafold2(nn.Module):
             C_mask = C_mask,
             cloud_mask = cloud_mask,
             bb_flat_mask_crossed = bb_flat_mask_crossed,
-            num_backbone_atoms = self.num_backbone_atoms
+            atom_mask = self.atom_mask
         )
 
-        # derive nodes
+        if not self.refine_coords or not refine: 
+            return coords
 
-        num_atoms_per_residue = coords.shape[-2] // seq.shape[1]
+        # derive nodes
 
         structure_embed = self.trunk_to_structure_dim(trunk_embeds)
         x = reduce(structure_embed, 'b i j d -> b i d', 'mean')
         x += self.structure_module_embeds(seq)
-        x = repeat(x, 'b n d -> b n l d', l = num_atoms_per_residue)
+        x = repeat(x, 'b n d -> b n l d', l = constants.NUM_COORDS_PER_RES)
 
         atom_tokens = scn_atom_embedd(seq)
         x += self.atom_tokens_embed(atom_tokens)
@@ -1325,8 +1377,7 @@ class Alphafold2(nn.Module):
         if exists(self.to_equivariant_net_edges):
             edges = self.to_equivariant_net_edges(trunk_embeds)
             edges = fourier_encode(edges, num_encodings = self.se3_edges_fourier_encodings, include_self = True)
-            edges = repeat(edges, 'b i j d -> b (i l) (j m) d', l = num_atoms_per_residue, m = num_atoms)
-            edges = edges.double()
+            edges = repeat(edges, 'b i j d -> b (i l1) (j l2) d', l1 = constants.NUM_COORDS_PER_RES, l2 = constants.NUM_COORDS_PER_RES)
 
         # derive adjacency matrix
         # todo - fix so Cbeta is connected correctly
@@ -1348,10 +1399,24 @@ class Alphafold2(nn.Module):
             pooled_feats = pooled_feats.double()
             structure_kwargs = {'global_feats': pooled_feats}
 
+        # prepare only atoms defined by atom mask
+
+        atom_mask = repeat(self.atom_mask, 'a -> () (n a)', n = n)
+        atom_mask_crossed = atom_mask[:, :, None] & atom_mask[:, None, :]
+        total_atoms = self.num_atoms * n
+
+        coords  = coords.masked_select(atom_mask[..., None]).reshape(b, total_atoms, -1)
+        x       = x.masked_select(atom_mask[..., None]).reshape(b, total_atoms, -1)
+        adj_mat = adj_mat.masked_select(atom_mask_crossed).reshape(b, total_atoms, total_atoms)
+        flat_chain_mask = flat_chain_mask.masked_select(atom_mask).reshape(b, total_atoms)
+
+        if exists(edges):
+            edges   = edges.masked_select(atom_mask_crossed[..., None]).reshape(b, total_atoms, total_atoms, -1)
+
         # prepare float64 precision for equivariance
 
         original_dtype = coords.dtype
-        x, coords = map(lambda t: t.double(), (x, coords))
+        x, coords, edges = map(maybe(lambda t: t.double()), (x, coords, edges))
 
         # iterative refinement with equivariant transformer in high precision
 
